import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from collections import OrderedDict
from keras.models import load_model
import matplotlib
import matplotlib.animation as animation

from astrorapid.process_light_curves import read_multiple_light_curves, prepare_input_arrays

plt.rcParams['text.usetex'] = True
plt.rcParams['font.serif'] = ['Computer Modern Roman'] + plt.rcParams['font.serif']

CLASS_NAMES = ['Pre-explosion', 'SNIa-norm', 'SNIbc', 'SNII', 'SNIa-91bg', 'SNIa-x', 'Class A', 'Kilonova', 'SLSN-I',
               'Class B', 'Class C', 'Class D', 'TDE']
CLASS_COLOR = {'Pre-explosion': 'grey', 'SNIa-norm': 'tab:green', 'SNIbc': 'tab:orange', 'SNII': 'tab:blue',
               'SNIa-91bg': 'tab:red', 'SNIa-x': 'tab:purple', 'Class A': 'tab:brown', 'Kilonova': '#aaffc3',
               'SLSN-I': 'tab:olive', 'Class B': 'tab:cyan', 'Class C': '#FF1493', 'Class D': 'navy', 'TDE': 'tab:pink'}
PB_COLOR = {'u': 'tab:blue', 'g': 'tab:blue', 'r': 'tab:orange', 'i': 'm', 'z': 'k', 'Y': 'y'}
PB_MARKER = {'g': 'o', 'r': 's'}
PB_ALPHA = {'g': 0.3, 'r': 1.}


class Classify(object):
    def __init__(self, light_curves, model_filepath='keras_model.hdf5', passbands=('g', 'r')):
        """
        Takes a list of photometric information and classifies light curves as a function of time

        Parameters
        ----------
        light_curves : list
            Is a list of tuples. Each tuple contains the light curve information of a transient object in the form
            (mjd, flux, fluxerr, passband, zeropoint, photflag, ra, dec, objid, redshift, mwebv).
            Here, mjd, flux, fluxerr, passband, zeropoint, and photflag are arrays.
            ra, dec, objid, redshift, and mwebv are floats
        model_filepath : str
            Optional argument. The model is taken from the pre-trained model ZTF model if not specified.
        passbands : tuple
            Optional argument. A tuple listing each passband. E.g. ('g', 'r').

        """
        self.light_curves = light_curves
        self.model_filepath = model_filepath
        self.passbands = passbands
        self.model = load_model(self.model_filepath)

    def process_light_curves(self):
        processed_lightcurves = read_multiple_light_curves(self.light_curves)
        X = prepare_input_arrays(processed_lightcurves, self.passbands, contextual_info=(0,))

        return X

    def get_predictions(self):
        self.X, self.orig_lc, self.timesX, self.objids = self.process_light_curves()

        self.y_predict = self.model.predict(self.X)

        return self.y_predict

    def plot_light_curves_and_classifications(self, indexes_to_plot=None):
        """
        Plot light curve (top panel) and classifications (bottom panel) vs time.

        Parameters
        ----------
        indexes_to_plot : tuple
            The indexes listed in the tuple will be plotted according to the order of the input light curves.
            E.g. (0, 1, 3, 5) will plot the zeroth, first, third and fifth light curves and classifications.
            If None or True, then all light curves will be plotted

        """

        font = {'family': 'normal',
                'size': 33}
        matplotlib.rc('font', **font)

        if indexes_to_plot is None or indexes_to_plot is True:
            indexes_to_plot = np.arange(len(self.y_predict))

        if not hasattr(self, 'y_predict'):
            self.get_predictions()

        for idx in indexes_to_plot:
            argmax = self.timesX[idx].argmax() + 1
            fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, figsize=(13, 15),
                                           num="classification_vs_time_{}".format(idx), sharex=True)
            ax1.axvline(x=0, color='k', linestyle='-', linewidth=1)
            ax2.axvline(x=0, color='k', linestyle='-', linewidth=1)

            for pb in self.passbands:
                if pb in self.orig_lc[idx].keys():
                    ax1.errorbar(self.orig_lc[idx][pb]['time'], self.orig_lc[idx][pb]['fluxRenorm'],
                                 yerr=self.orig_lc[idx][pb]['fluxErrRenorm'], fmt=PB_MARKER[pb], label=pb,
                                 c=PB_COLOR[pb], lw=3, markersize=10)

            for classnum, classname in enumerate(CLASS_NAMES):
                ax2.plot(self.timesX[idx][:argmax], self.y_predict[idx][:, classnum][:argmax], '-', label=classname,
                         color=CLASS_COLOR[classname], linewidth=3)
            ax1.legend(frameon=False, fontsize=33)
            ax2.legend(frameon=False, fontsize=23.5)  # , ncol=2)
            ax2.set_xlabel("Days since trigger (rest frame)")  # , fontsize=18)
            ax1.set_ylabel("Relative Flux")  # , fontsize=15)
            ax2.set_ylabel("Class Probability")  # , fontsize=18)
            ax1.set_ylim(-0.1, 1.1)
            ax2.set_ylim(0, 1)
            ax1.set_xlim(-70, 80)
            plt.setp(ax1.get_xticklabels(), visible=False)
            ax2.yaxis.set_major_locator(MaxNLocator(nbins=6, prune='upper'))  # added
            plt.tight_layout()
            fig.subplots_adjust(hspace=0)
            plt.savefig('classification_vs_time_{}.pdf'.format(self.objids[idx]))
            plt.close()

    def plot_classification_animation(self, indexes_to_plot=None):
        """
        Plot light curve (top panel) and classifications (bottom panel) vs time as an mp4 animation.

        Parameters
        ----------
        indexes_to_plot : tuple
            The indexes listed in the tuple will be plotted according to the order of the input light curves.
            E.g. (0, 1, 3, 5) will plot the zeroth, first, third and fifth light curves and classifications.
            If None or True, then all light curves will be plotted

        """

        font = {'family': 'normal',
                'size': 33}
        matplotlib.rc('font', **font)

        if indexes_to_plot is None or indexes_to_plot is True:
            indexes_to_plot = np.arange(len(self.y_predict))

        if not hasattr(self, 'y_predict'):
            self.get_predictions()

        for idx in indexes_to_plot:
            argmax = self.timesX[idx].argmax() + 1
            fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, figsize=(13, 15),
                                           num="animation_classification_vs_time_{}".format(idx), sharex=True)

            ax1.legend(frameon=False, fontsize=33)
            ax2.legend(frameon=False, fontsize=23.5)  # , ncol=2)
            ax2.set_xlabel("Days since trigger (rest frame)")  # , fontsize=18)
            ax1.set_ylabel("Relative Flux")  # , fontsize=15)
            ax2.set_ylabel("Class Probability")  # , fontsize=18)
            ax1.set_ylim(-0.1, 1.1)
            ax2.set_ylim(0, 1)
            ax1.set_xlim(-70, 80)
            plt.setp(ax1.get_xticklabels(), visible=False)
            ax2.yaxis.set_major_locator(MaxNLocator(nbins=6, prune='upper'))  # added
            plt.tight_layout()
            fig.subplots_adjust(hspace=0)
            ax1.axvline(x=0, color='k', linestyle='-', linewidth=1)
            ax2.axvline(x=0, color='k', linestyle='-', linewidth=1)

            Writer = animation.writers['ffmpeg']
            writer = Writer(fps=5, bitrate=1800)

            def animate(i):
                for pbidx, pb in enumerate(self.passbands):
                    ax1.plot(self.timesX[idx][:argmax][:int(i + 1)], self.X[idx][:, pbidx][:argmax][:int(i + 1)],
                             label=pb, c=PB_COLOR[pb], lw=3)  # , markersize=10, marker=MARKPB[pb])

                for classnum, classname in enumerate(CLASS_NAMES):
                    ax2.plot(self.timesX[idx][:argmax][:int(i + 1)], self.y_predict[idx][:, classnum][:argmax][:int(i + 1)],
                             '-', label=classname, color=CLASS_COLOR[classname], linewidth=3)

                # Don't repeat legend items
                ax1.legend(frameon=False, fontsize=33)
                ax2.legend(frameon=False, fontsize=23.5)
                handles1, labels1 = ax1.get_legend_handles_labels()
                handles2, labels2 = ax2.get_legend_handles_labels()
                by_label1 = OrderedDict(zip(labels1, handles1))
                by_label2 = OrderedDict(zip(labels2, handles2))
                ax1.legend(by_label1.values(), by_label1.keys(), frameon=False, fontsize=33)
                ax2.legend(by_label2.values(), by_label2.keys(), frameon=False, fontsize=23.5)

            ani = animation.FuncAnimation(fig, animate, frames=50, repeat=True)
            ani.save(os.path.join('classification_vs_time_{}.mp4'.format(self.objids[idx])), writer=writer)



# -*- coding: utf-8 -*-
"""
Created on Thu Jan 12 14:30:12 2017

@author: danielgodinez
"""
import numpy as np
from astropy.stats import median_absolute_deviation
from scipy.integrate import quad
from scipy.cluster.hierarchy import fclusterdata

def shannon_entropy(mag, magerr):
    """Shannon entropy (Shannon et al. 1949) is used as a metric to quantify the amount of
    information carried by a signal. The procedure employed here follows that outlined by
    (D. Mislis et al. 2015). The probability of each point is given by a Cumulative Distribution 
    Function (CDF). Following the same procedure as (D. Mislis et al. 2015), this function employs
    both the normal and inversed gaussian CDF, with the total shannon entropy given by a combination of
    the two. See: (SIDRA: a blind algorithm for signal detection in photometric surveys, D. Mislis et al., 2015)
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.
    
    :rtype: float
    """
    
    
    mag, magerr = remove_bad(mag, magerr)
    
    mean =  meanMag(mag, magerr)   
    RMS = RootMS(mag, magerr)
    
    p_list1 = []
    p_list2 = [] 
    inv_list1 = []
    inv_list2 = []
    
    t = range(0, len(mag))
    d_delta = [i*2.0 for i in magerr]
    
    """Error fn definition: http://mathworld.wolfram.com/Erf.html"""
    def errfn(x):
        def integral(t):
            integrand =  (2./np.sqrt(np.pi))*np.e**(-t**2)
            return integrand 
        integ, err = quad(integral, 0, x)
        return integ
        
    """The Gaussian CDF: http://mathworld.wolfram.com/NormalDistribution.html"""   
    def normal_gauss(x):
        return 0.5*(1. + errfn(x))
    
    """Inverse Gaussian CDF: http://mathworld.wolfram.com/InverseGaussianDistribution.html"""    
    def inv_gauss(x, y):
        return 0.5*(1. + errfn(x)) + (0.5*np.e**((2.*RMS)/mean))*(1. - errfn(y))
        
    def shannon_entropy1(mag, magerr):
        """
        This function utilizes the normal Gaussian CDF to set the probability of 
        each point in the lightcurve and computes the Shannon Entropy given this distribution.
        """
        
        for i in t:
            val = normal_gauss((mag[i] + magerr[i] - mean)/(RMS*np.sqrt(2)))
            p_list1.append(val)
        
            val2 = normal_gauss((mag[i] - magerr[i] - mean)/(RMS*np.sqrt(2)))
            p_list2.append(val2)
                                    
        p_list3 = [1 if i <= 0 else i for i in p_list1]
        p_list4 = [1 if i <= 0 else i for i in p_list2]
        
        entropy = -sum(np.log2(p_list3)*d_delta + np.log2(p_list4)*d_delta)
        return entropy
            
    def shannon_entropy2(mag, magerr):
        """
        This function utilizes the inverse Gaussian CDF to set the probability of each point
        in the lightcurve and computes the Shannon Entropy given this distribution.
        """
        
        for i in t:
            val = inv_gauss(np.sqrt(RMS/(2.*(mag[i] + magerr[i])))*(((mag[i] + magerr[i])/mean) - 1.), 
                            np.sqrt(RMS/(2.*(mag[i] + magerr[i])))*(((mag[i] + magerr[i])/mean) + 1.))
            inv_list1.append(val)
            
            val2 = inv_gauss(np.sqrt(RMS/(2.*(mag[i] - magerr[i])))*(((mag[i] - magerr[i])/mean) - 1.), 
                            np.sqrt(RMS/(2.*(mag[i] - magerr[i])))*(((mag[i] - magerr[i])/mean) + 1.))
            inv_list2.append(val2)         
            
        inv_list3 = [1 if i <= 0 else i for i in inv_list1]
        inv_list4 = [1 if i <= 0 else i for i in inv_list2]
        
        entropy = -sum(np.log2(inv_list3)*d_delta + np.log2(inv_list4)*d_delta)
        return entropy
        
    """The total Shannon Entropy is calculated by adding the values calculated using both the normal
    and inverse Gaussian CDF
    """      
    total_entropy = np.nan_to_num(shannon_entropy1(mag, magerr) + shannon_entropy2(mag, magerr))
    return total_entropy

   
def auto_correlation(mag, magerr):
    """The autocorrelation integral calculates the correlation of a given signal as a function of 
    the time delay of each measurement. Has been employed in previous research as a metric to 
    differentitate between lightcurve classes. See: (SIDRA: a blind algorithm for signal
    detection in photometric surveys, D. Mislis et al., 2015)
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.  
    :param mag: the time-varying intensity of the lightcurve. Must be an array.

    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)    
    n = np.float(len(mag))
    mean = np.median(mag)
    RMS = RootMS(mag, magerr)
    t = range(1, len(mag))
        
    sum_list = []
    val_list = []
    
            
    for i in t:
        sum1 = np.array(sum((mag[0:int(n)-i] - mean)*(mag[i:int(n)] - mean)))
        sum_list.append(sum1)
        
        val = np.array(1/((n-i)*RMS**2))
        val_list.append(val)
        
    auto_corr = abs(sum([x*y for x,y in zip(sum_list, val_list)]))        
   
    return auto_corr
"""
def con(mag, magerr):



    
    
    std = deviation(mag, magerr)
    
    
    deviatingThreshold = median - magerr
    con = 0  
    deviating = False
    
    a = np.argwhere(magerr < deviatingThreshold)
    
    if len(a) < 3:
        return 0
    else:
        for i in xrange(len(magerr)-2):
            first = magerr[i]
            second = magerr[i+1]
            third = magerr[i+2]
            if (first <= deviatingThreshold and
                second <= deviatingThreshold and
                third <= deviatingThreshold):
                    if (not deviating):
                        con += 1
                        deviating = True
                    elif deviating:
                        deviating = False


    con2 = 0
    deviating = False
    a = np.argwhere(magerr < deviatingThreshold)
    deviatingThreshold = median + magerr

    if len(a) < 3:
        return 0
    else:
        for i in xrange(len(mag)-2):
            first = mag[i]
            second = mag[i+1]
            third = mag[i+2]
            if (first <= deviatingThreshold and
                second <= deviatingThreshold and
                third <= deviatingThreshold):
                    if (not deviating):
                        con += 1
                        deviating = True
                    elif deviating:
                        deviating = False

    return con + con2


def con2(mag, magerr):

    
    mag, magerr = remove_bad(mag, magerr)
    diff = mag - meanMag(mag, magerr)
    hist, edges = np.histogram(diff, bins = 10)
    val = np.where(hist == max(hist))
    bin_range = np.where((diff > edges[val[0][0]]) & (diff < edges[val[0][0]+1]))
    
    mean = meanMag(mag[bin_range], magerr[bin_range])
    std = deviation(mag, magerr)    
    deviatingThreshold = mean - 2*std    
    con = 0  
    deviating = False
    
    a = np.argwhere(magerr < deviatingThreshold)
    
    if len(a) < 3:
        return 0
    else:
        for i in xrange(len(mag)-2):
            first = mag[i]
            second = mag[i+1]
            third = mag[i+2]    
            if (first <= deviatingThreshold and
                second <= deviatingThreshold and
                third <= deviatingThreshold):
                    if (not deviating):
                        con += 1
                        deviating = True
                    elif deviating:
                        deviating = False

    return con
"""

def con(mag, magerr):
    """Con is defined as the number of clusters containing three or more
        consecutive observations with magnitudes brighter than the median
        magnitude plus 3 standard deviations. For a microlensing event Con = 1,
        assuming a  flat lightcurve prior to the event.
        :param mag: the time-varying intensity of the lightcurve. Must be an array.
        :param magerr: photometric error for the intensity. Must be an array.
        
        :rtype: float
    """
    
    #diff = mag - meanMag(mag, magerr)
    #hist, edges = np.histogram(diff, bins = 10)
    #val = np.where(hist == max(hist))
    #bin_range = np.where((diff > edges[val[0][0]]) & (diff < edges[val[0][0]+1]))
    
    mag, magerr = remove_bad(mag, magerr)
    median = np.median(mag)
    
    deviatingThreshold = np.array(median - magerr)
    con = 0
    deviating = False
    
    a = np.argwhere(magerr < deviatingThreshold)
    
    if len(a) < 3:
        return 0
    else:
        for i in xrange(len(magerr)-2):
            first = mag[i]
            second = mag[i+1]
            third = mag[i+2]
            if ((first <= deviatingThreshold[i]).all() == True and
                (second <= deviatingThreshold[i+1]).all() == True and
                (third <= deviatingThreshold[i+2]).all() == True):
                if (not deviating):
                    con += 1
                    deviating = True
                elif deviating:
                    deviating = False

    con2 = 0
    deviating = False

    deviatingThreshold = np.array(median + magerr)
    a = np.argwhere(magerr > deviatingThreshold)
    
    if len(a) < 3:
        return 0
    else:
        for i in xrange(len(mag)-2):
            first = mag[i]
            second = mag[i+1]
            third = mag[i+2]
            if ((first >= deviatingThreshold).all() == True and
                (second >= deviatingThreshold).all() == True and
                (third >= deviatingThreshold).all() == True):
                if (not deviating):
                    con2 += 1
                    deviating = True
                elif deviating:
                    deviating = False

    return con + con2


def con2(mag, magerr):
    """Only looks at bin below the reference magnitude
    """

    mag, magerr = remove_bad(mag, magerr)
    median = np.median(mag)
    
    deviatingThreshold = np.array(median + magerr)
    con = 0
    deviating = False
    
    a = np.argwhere(magerr > deviatingThreshold)
    
    if len(a) < 3:
        return 0
    else:
        for i in xrange(len(mag)-2):
            first = mag[i]
            second = mag[i+1]
            third = mag[i+2]
            if ((first >= deviatingThreshold).all() == True and
                (second >= deviatingThreshold).all() == True and
                (third >= deviatingThreshold).all() == True):
                if (not deviating):
                    con += 1
                    deviating = True
                elif deviating:
                    deviating = False

    return con


def kurtosis(mag, magerr):
    """"Kurtosis function returns the calculated kurtosis of the lightcurve. 
    It's a measure of the peakedness (or flatness) of the lightcurve relative 
    to a normal distribution. See: www.xycoon.com/peakedness_small_sample_test_1.htm
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.

    :rtype: float
    """""
    
    mag, magerr = remove_bad(mag, magerr)
    #mean = meanMag(mag, magerr)
    mean = np.median(mag)
    std = deviation(mag, magerr)
   
    n = np.float(len(mag))
    kurtosis = (n*(n+1.)/((n-1.)*(n-2.)*(n-3.))*sum(((mag - mean)/std)**4)) - \
    (3.*((n-1.)**2)/((n-2.)*(n-3.)))
    return kurtosis
        

def skewness(mag, magerr):
    """Skewness measures the assymetry of a lightcurve, with a positive skewness
    indicating a skew to the right, and a negative skewness indicating a skew to the left. 
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.

    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    #mean = meanMag(mag, magerr)
    mean = np.median(mag)
    std = deviation(mag, magerr)
    n = np.float(len(mag))
    
    skewness = (1./n)*sum((mag - mean)**3/std**3)
    #skewness = skew(mag, axis = 0, bias = True)
    return skewness

def vonNeumannRatio(mag, magerr):
    """The von Neumann ratio η was defined in 1941 by John von Neumann and serves as the 
    mean square successive difference divided by the sample variance. When this ratio is small, 
    it is an indication of a strong positive correlation between the successive photometric data points. 
    See: (J. Von Neumann, The Annals of Mathematical Statistics 12, 367 (1941))
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.

    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    n = np.float(len(mag))
    delta = sum((mag[1:] - mag[:-1])**2 / (n-1.))
    sample_variance = deviation(mag, magerr)**2
    
    vonNeumannRatio = delta / sample_variance
    return vonNeumannRatio
    
def stetsonJ(mag, magerr):
    """The variability index K was first suggested by Peter B. Stetson and serves as a 
    measure of the kurtosis of the magnitude distribution. 
    See: (P. B. Stetson, Publications of the Astronomical Society of the Pacific 108, 851 (1996)).
        
    :param mag: the time-varying intensity of the object. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.
    
    :rtype: float
    """

    mag, magerr = remove_bad(mag, magerr)
    n = np.float(len(mag))
    #mean = meanMag(mag, magerr)
    mean = np.median(mag)
    delta_list = []
    
    for i in range(0, len(mag)):
        delta = np.sqrt(n/(n-1.))*((mag[i] - mean)/magerr[i])
        delta_list.append(delta)
    
    val = np.nan_to_num([x*y for x,y in zip(delta_list[0:int(n)-1], delta_list[1:int(n)])])    
    sign = np.sign(val)
    stetj = sum(sign*np.sqrt(np.abs(val)))
    return stetj
  
def stetsonK(mag, magerr):
    """The variability index K was first suggested by Peter B. Stetson and serves as a 
    measure of the kurtosis of the magnitude distribution. 
    See: (P. B. Stetson, Publications of the Astronomical Society of the Pacific 108, 851 (1996)).
        
    :param mag: the time-varying intensity of the object. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.
    
    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)            
    n = np.float(len(mag))
    #mean = meanMag(mag, magerr)
    mean = np.median(mag)
    
    delta = np.sqrt((n/(n-1.)))*((mag - mean)/magerr)
        
    stetsonK = ((1./n)*sum(abs(delta)))/(np.sqrt((1./n)*sum(delta**2)))
    return np.nan_to_num(stetsonK)


    
def median_buffer_range(mag, magerr):
    """This function returns the ratio of points that are between plus or minus 10% of the
    amplitude value over the mean
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.
    
    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    n = np.float(len(mag))
    amp = amplitude(mag, magerr) 
    #mean = meanMag(mag, magerr)
    mean = np.median(mag)
    a = mean - amp/10. 
    b = mean + amp/10. 
    
    median_buffer_range = len(np.argwhere((mag > a) & (mag < b))) / n
    
    return median_buffer_range

def median_buffer_range2(mag, magerr):
    """This function returns the ratio of points that are more than 20% of the amplitude
    value over the mean
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.

    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    n = np.float(len(mag))
    amp = amplitude(mag, magerr) 
    #mean = meanMag(mag, magerr)
    mean = np.median(mag)
    a = mean - amp/5. 
 
    
    median_buffer_range = len(np.argwhere((mag < a))) / n
    
    return median_buffer_range
    
    
def std_over_mean(mag, magerr):
    """A measure of the ratio of standard deviation and mean, both weighted by the errors.
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.

    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    mean = meanMag(mag, magerr)
    std = deviation(mag, magerr)
    
    std_over_mean = std/mean
    return std_over_mean
      
    
def amplitude(mag, magerr):
    """The amplitude of the lightcurve defined as the difference between the maximum magnitude
    measurement and the lowest magnitude measurement. To account for outliers, an array of the
    absolute value of the magnitude minus weighted mean is created. From this array, a 5% 
    threshold is applied such that top 5% of points are ommitted as outliers and the amplitude
    is left to be defined as the maximun magnitude minus the minimum magnitude of the remaining points. 
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.
    
    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    array_list = [] 
    
    for i in range(0, len(mag)):
        array = np.abs(mag[i] -np.median(mag))
        array_list.append(array)
        
    index = np.argsort(array_list)
    index2 = len(mag) - np.int(len(mag)*0.95)
    threshold = index[-index2]
    
    mag = np.delete(mag, threshold)
    amplitude = np.max(mag) - np.min(mag)
    
    return amplitude

def median_distance(mjd, mag, magerr):

    bad = np.where(mag == 0)
    magerr = np.delete(magerr, bad)
    mag = np.delete(mag, bad)
    mjd = np.delete(mjd, bad)
    
    bad = np.where(np.isnan(magerr) == True)
    magerr = np.delete(magerr, bad)
    mag = np.delete(mag, bad)
    mjd = np.delete(mjd, bad)
    
    bad = np.where(np.isfinite == False)
    magerr = np.delete(magerr, bad)
    mag = np.delete(mag, bad)
    mjd = np.delete(mjd, bad)
    
    mag1 = (mag[1:] - mag[:-1])**2
    time1 = (mjd[1:] - mjd[:-1])**2

    distance = np.median(np.sqrt(mag1 + time1))
    return distance

def clusters(mag, magerr):
    
    mag, magerr = remove_bad(mag, magerr)
    new_mag = mag.reshape(len(mag), 1)

    cluster = fclusterdata(new_mag, 0.1, criterion='distance')
    num_clusters =  len(np.unique(cluster))
    return num_clusters

def above1(mag, magerr):
    """This function measures the ratio of data points that are above 1 standard deviation 
    from the mean magnitude.
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.

    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    a = meanMag(mag, magerr) - deviation(mag, magerr)
    
    above1 = len(np.argwhere(mag < a) )  
    
    return above1

def above3(mag, magerr):
    """This function measures the ratio of data points that are above 3 standard deviations 
    from the mean magnitude.
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.
    
    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    a = meanMag(mag, magerr) - 3*deviation(mag, magerr)
    
    above3 = len(np.argwhere(mag < a) )  
    
    return above3

def above5(mag, magerr):
    """This function measures the ratio of data points that are above 5 standard deviations
    from the mean magnitude.
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.
    
    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    a = meanMag(mag, magerr) - 5*deviation(mag, magerr)
    
    above5 = len(np.argwhere(mag < a))    
    
    return above5
        
def below1(mag, magerr):
    """This function measures the ratio of data points that are below 1 standard deviations 
    from the mean magnitude.
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.

    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    a = meanMag(mag, magerr) + deviation(mag, magerr)
    
    below1 = len(np.argwhere(mag > a)) 
    
    return below1
    
def below3(mag, magerr):
    """This function measures the ratio of data points that are below 3 standard deviations
    from the mean magnitude.
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.
    
    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    a = meanMag(mag, magerr) + 3*deviation(mag, magerr)
    
    below3 = len(np.argwhere(mag > a))
    
    return below3
    
def below5(mag, magerr):
    """This function measures the ratio of data points that are below 5 standard deviations
    from the mean magnitude.
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.
    
    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    a = meanMag(mag, magerr) + 5*deviation(mag, magerr)
    
    below5 = len(np.argwhere(mag > a))  
    
    return below5
    
def medianAbsDev(mag, magerr):
    """"A measure of the mean average distance between each magnitude value 
    and the mean magnitude.
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.

    :rtype: float
    """
    
    mag = remove_bad(mag, magerr)[0]
    medianAbsDev = median_absolute_deviation(mag)
    
    return medianAbsDev
    
def RootMS(mag, magerr):
    """A measure of the root mean square deviation, weighted by the errors.
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.

    :rtype: float
    """
    
    mag, magerr = remove_bad(mag, magerr)
    #mean = meanMag(mag, magerr)
    mean = np.median(mag)
    rms = np.sqrt(sum(((mag - mean)/magerr)**2)/sum(1./magerr**2))
    
    return rms

    
def meanMag(mag, magerr):
    """Calculates mean magnitude, weighted by the errors.
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.
    
    rtype: float
    """
   
    mag, magerr = remove_bad(mag, magerr)
    mean = sum(mag/magerr**2)/sum(1./magerr**2)
    
    return mean
    
def deviation(mag, magerr):
     """Calculates the standard deviation, weighted by the errors.
    
     :param mag: the time-varying intensity of the lightcurve. Must be an array.  
     :param magerr: photometric error for the intensity. Must be an array.
    
     rtype: float
     """

     mag, magerr = remove_bad(mag, magerr)
     m = np.float(len(np.argwhere(magerr != 0)))
     std = np.sqrt(sum((1./magerr**2)*(mag - meanMag(mag, magerr))**2)/(((m - 1.)/m)*sum(1./magerr**2)))
     
     return std
        
def remove_bad(mag, magerr):
    """This function removes all points in the lightcurve that do not contain
    a calculation for photometric error.
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.
    
    rtype: float
    """
    
    bad = np.where(magerr == 0)
    magerr = np.delete(magerr, bad)
    mag = np.delete(mag, bad)
    
    bad = np.where(mag == 0)
    magerr = np.delete(magerr, bad)
    mag = np.delete(mag, bad)
    
    bad = np.where(np.isnan(magerr) == True)
    magerr = np.delete(magerr, bad)
    mag = np.delete(mag, bad)
    
    bad = np.where(np.isfinite(magerr) == False)
    magerr = np.delete(magerr, bad)
    mag = np.delete(mag, bad)
    
    return mag, magerr

def beyond1std(mag, magerr):
    
    num = above1(mag, magerr)
    above1_std = np.float(num/len(mag))
    
    num2 = below1(mag, magerr)
    below1_std = np.float(num2/len(mag))
    tot_beyond = below1_std + above1_std
    
    return tot_beyond
        
def compute_statistics(mjd, mag, magerr):
    """This function will compute all the statistics and return them in an array in the 
    following order: shannon_entropy, auto_correlation, kurtosis, skewness, vonNeumannRatio,
    stetsonJ, stetsonK, median_buffer_Rance, std_over_mean, std_over_mean, below1, medianAbdsDev, RMS
    
    :param mag: the time-varying intensity of the lightcurve. Must be an array.
    :param magerr: photometric error for the intensity. Must be an array.
    
    :return: the function will return an array with the statistics.
    :rtype: array, float
    """
    
    stat_array = (shannon_entropy(mag, magerr), auto_correlation(mag, magerr), kurtosis(mag, magerr),
             skewness(mag, magerr), vonNeumannRatio(mag, magerr), stetsonJ(mag, magerr), stetsonK(mag, magerr), con(mag, magerr),
  median_buffer_range(mag, magerr), median_buffer_range2(mag, magerr), std_over_mean(mag, magerr),
                medianAbsDev(mag, magerr), RootMS(mag, magerr), amplitude(mag, magerr), median_distance(mjd, mag, magerr), clusters(mag, magerr), deviation(mag, magerr), beyond1std(mag, magerr), con(mag, magerr), con2(mag, magerr))
    
    #stat_array = (auto_correlation(mag, magerr), kurtosis(mag, magerr), skewness(mag, magerr), vonNeumannRatio(mag, magerr),
    #               stetsonJ(mag, magerr), stetsonK(mag, magerr), median_buffer_range(mag, magerr),
    #              median_buffer_range2(mag, magerr), std_over_mean(mag, magerr), medianAbsDev(mag, magerr), RootMS(mag, magerr))
  
    return stat_array



from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.cosmology import WMAP9 as cosmo
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

import helpers
from ANTARES_object.LAobject import LAobject


class InputLightCurve(object):
    def __init__(self, mjd, flux, fluxerr, passband, zeropoint, photflag, ra, dec, objid, redshift=None, mwebv=None):
        """

        Parameters
        ----------
        mjd : array
            List of floating point mjd times for entire light curve in all passbands.
        flux : array
            List of floating point fluxes for entire light curve in all passbands.
        fluxerr : array
            List of floating point flux errors times for entire light curve in all passbands.
        passband : array
            List of passbands as strings for each point in the array.
        zeropoint : array
            List of floating point zeropoints for entire light curve in all passbands.
        photflag : array
            List of integer flags indicating whether the observation was a detection or non-detection for each
            point in the light curve.
        ra : float
            Right Ascension in degrees.
        dec : float
            Declination in degrees.
        objid : str
            String to identify the input object.
        redshift : float
            Optional parameter. Cosmological redshift.
        mwebv : float
            Optional parameter. Milky Way E(B-V) extinction.
        """

        self.mjd = np.array(mjd)
        self.flux = np.array(flux)
        self.fluxerr = np.array(fluxerr)
        self.passband = np.array(passband)
        self.zeropoint = np.array(zeropoint)
        self.photflag = np.array(photflag)
        self.ra = ra
        self.dec = dec
        self.objid = objid
        self.redshift = redshift
        self.mwebv = mwebv

        self.b = self.get_galactic_latitude()
        self.trigger_mjd, self.t = self.get_trigger_time()


    def get_galactic_latitude(self):
        c_icrs = SkyCoord(ra=self.ra * u.degree, dec=self.dec * u.degree, frame='icrs')
        b = c_icrs.galactic.b.value

        return b

    def get_trigger_time(self):
        trigger_mjd = float(self.mjd[self.photflag == 6144][0])
        t = self.mjd - trigger_mjd

        return trigger_mjd, t

    def correct_time_dilation(self, t):
        t = t / (1 + self.redshift)

        return t

    def correct_for_distance(self, flux, fluxerr):
        dlmu = cosmo.distmod(self.redshift).value
        flux, fluxerr = helpers.calc_luminosity(flux, fluxerr, dlmu)

        return flux, fluxerr

    def preprocess_light_curve(self):
        """ Preprocess light curve. """

        # Account for distance and time dilation if redshift is known
        if self.redshift is not None:
            self.t = self.correct_time_dilation(self.t)
            self.flux, self.fluxerr = self.correct_for_distance(self.flux, self.fluxerr)

        obsid = np.arange(len(self.t))

        laobject = LAobject(locusId=self.objid, objectId=self.objid, time=self.t, flux=self.flux, fluxErr=self.fluxerr,
                            obsId=obsid, passband=self.passband, zeropoint=self.zeropoint, per=False, mag=False,
                            photflag=self.photflag, z=self.redshift)

        outlc = laobject.get_lc(recompute=True)

        otherinfo = [self.redshift, self.b, self.mwebv, self.trigger_mjd, self.objid]

        savepd = {
        pb: pd.DataFrame(lcinfo).loc[[0, 5, 6, 7]].rename({0: 'time', 5: 'fluxRenorm', 6: 'fluxErrRenorm', 7: 'photflag'}).T
        for pb, lcinfo in
        outlc.items()}  # Convert to dataframe rows: time, fluxNorm, fluxNormErr, photFlag; columns: ugrizY
        savepd['otherinfo'] = pd.DataFrame(otherinfo)
        savepd = pd.DataFrame(
            {(outerKey, innerKey): values for outerKey, innerDict in savepd.items() for innerKey, values in
             innerDict.items()})  # Use multilevel indexing

        return savepd


def read_multiple_light_curves(light_curve_list):
    """
    light_curve_list is a list of tuples with each tuple having entries:
    mjd, flux, fluxerr, passband, zeropoint, photflag, ra, dec, objid, redshift, mwebv

    Returns the processed light curves
    """

    processed_light_curves = []
    for light_curve in light_curve_list:
        inputlightcurve = InputLightCurve(*light_curve)
        processed_light_curves.append(inputlightcurve.preprocess_light_curve())

    return processed_light_curves


def prepare_input_arrays(lightcurves, passbands=('g', 'r'), contextual_info=(0,)):

    nobjects = len(lightcurves)
    nobs = 50
    npassbands = 2
    nfeatures = npassbands + 1

    X = np.zeros(shape=(nobjects, nfeatures, nobs))
    timesX = np.zeros(shape=(nobjects, nobs))
    objids_list = []
    orig_lc = []

    for i, data in enumerate(lightcurves):
        objid = i
        print("Preparing light curve {} of {}".format(i, nobjects))

        otherinfo = data['otherinfo'].values.flatten()
        redshift, b, mwebv, trigger_mjd, objid = otherinfo[0:5]

        # Make cuts
        if abs(b) < 15:
            print("In galactic plane. b = {}".format(b))
        if data.shape[0] < 4:
            print("Less than 4 epochs. nobs = {}".format(data.shape))
        time = data['r']['time'][0:nobs].dropna()
        if len(time[time < 0]) < 3:
            print("Less than 3 points in the r band pre trigger", len(time[time < 0]))

        # Get min and max times for tinterp
        timestep = 3.0
        mintimes = []
        maxtimes = []
        for j, pb in enumerate(passbands):
            if pb not in data:
                continue
            time = data[pb]['time'][0:nobs].dropna()
            mintimes.append(time.min())
            maxtimes.append(time.max())
        if mintimes == []:
            print("No data for objid:", i)
            continue
        mintime = min(mintimes)
        maxtime = max(maxtimes) + timestep
        tinterp = np.arange(mintime, maxtime, step=timestep)
        len_t = len(tinterp)
        if len_t >= nobs:
            tinterp = tinterp[(tinterp >= -70) & (tinterp <= 80)]
            len_t = len(tinterp)
            if len_t >= nobs:
                tinterp = tinterp[tinterp <= 80]
                len_t = len(tinterp)
                if len_t >= nobs:
                    tinterp = tinterp[len_t - nobs:]
                    len_t = len(tinterp)

        timesX[i][0:len_t] = tinterp

        endtime = 1000
        j = 0

        orig_lc.append(data)
        objids_list.append(objid)

        for j, pb in enumerate(passbands):
            if pb not in data:
                print("No", pb, "in objid:", objid)
                continue
            time = data[pb]['time'][0:nobs].dropna()
            try:
                flux = data[pb]['fluxRenorm'][0:nobs].dropna()
                fluxerr = data[pb]['fluxErrRenorm'][0:nobs].dropna()
            except KeyError:
                flux = data[pb][5][0:nobs].dropna()
                fluxerr = data[pb][6][0:nobs].dropna()
            photflag = data[pb]['photflag'][0:nobs].dropna()

            n = len(flux)  # Get vector length (could be less than nobs)

            if n > 1:
                if flux.values[-1] > flux.values[-2]:  # If last values are increasing, then set fill_values to zero
                    f = interp1d(time, flux, kind='linear', bounds_error=False, fill_value=0.)
                else:
                    f = interp1d(time, flux, kind='linear', bounds_error=False,
                                 fill_value='extrapolate')  # extrapolate until all passbands finished.

                fluxinterp = f(tinterp)
                fluxinterp = np.nan_to_num(fluxinterp)
                fluxinterp = fluxinterp.clip(min=0)
                X[i][j][0:len_t] = fluxinterp

        # Add contextual information
        for jj, c_idx in enumerate(contextual_info, 1):
            try:
                X[i][j + jj][0:len_t] = otherinfo[c_idx] * np.ones(len_t)
            except Exception as e:
                X[i][j + jj][0:len_t] = otherinfo[c_idx].values[0] * np.ones(len_t)

        # Correct shape for keras is (N_objects, N_timesteps, N_passbands) (where N_timesteps is lookback time)
        X = X.swapaxes(2, 1)

        return X, orig_lc, timesX, objids_list



# -*- coding: UTF-8 -*-
"""
Functions for spline interpolation with LAobjectss

"""

from __future__ import absolute_import
from __future__ import unicode_literals
import numpy as np
import scipy.interpolate as scinterp

class SplineMixin(object):

    def spline_smooth(self, per=None, minobs=10, phase_offset=None, recompute=False, smoothed=False, gpr=True):
        """
        per = fit a periodic spline
        minobs = mininum number of observations in each filter before we fit
        """
        outtck = getattr(self, 'outtck', None)

        if outtck is not None:
            if not recompute:
                return outtck
        else:
            outtck = {}

        filters = self.filters
        outlc = self.get_lc(recompute=recompute, per=per, phase_offset=phase_offset, smoothed=smoothed, gpr=gpr)

        # stick to cubic interpolation
        k = 3

        for i, pb in enumerate(filters):
            thislc = outlc.get(pb)
            thisphase, thisFlux, thisFluxErr = thislc
            nobs = len(thisphase)
            if nobs < minobs:
                continue
            m2 = thisphase.argsort()
            minphase = thisphase[m2].min()
            maxphase = thisphase[m2].max()

            # don't try and fit pathalogical lightcurves which all have
            # different mags measured at the same time
            if minphase==maxphase:
                continue

            # degree can be cubic but will get clipped if needed
            degree = k

            x = thisphase[m2]
            y = thisFlux[m2]

            x, m2 = np.unique(x, return_index=True)
            y = y[m2]

            cv = np.array(list(zip(x,y)))
            count = len(cv)

            # If periodic, extend the point array by count+degree+1
            if per > 0:
                factor, fraction = divmod(count+degree+1, count)
                cv = np.concatenate((cv,) * factor + (cv[:fraction],))
                count = len(cv)
                degree = np.clip(degree,1,degree)
            # If opened, prevent degree from exceeding count-1
            else:
                degree = np.clip(degree,1,count-1)

            # Calculate knot vector
            kv = None
            if per > 0:
                kv = np.arange(0-degree,count+degree+degree-1,dtype='int')
            else:
                kv = np.array([0]*degree + list(range(count-degree+1)) + [count-degree]*degree,dtype='int')

            # Calculate query range
            periodic = int(bool(per))
            u = np.linspace(periodic,(count-degree),count)
            tck = (kv, cv, degree)

            outtck[pb] = tck, u
        self.outtck = outtck
        return outtck




# -*- coding: UTF-8 -*-
"""
Functions for the derivation of periodic features with LAobjects
"""

from __future__ import absolute_import
from __future__ import unicode_literals
import warnings
from gatspy import periodic


class PeriodicMixin(object):
    """
    Features computed for periodic events
    """

    def periodscan(self, min_p, max_p, periods=None):
        """
        Computes a LombScargle periodgram from minp to maxp, or alternatively
        on the array of periods sets model, periods, P_multi and best_period,
        overwriting if they already exist
        """
        nobs = len(self.time)
        if nobs == 0:
            return None, None, None

        # if self.time.max() - self.time.min() < max_p:
        #     return None, None, None

        optimizer_kwds = {'quiet':True}
        try:
            if nobs <= 100:
                model = periodic.LombScargleMultiband(fit_period=True, optimizer_kwds=optimizer_kwds)
            else:
                print('nobs', nobs, self.objectId)
                model = periodic.LombScargleMultibandFast(fit_period=True, optimizer_kwds=optimizer_kwds)
            model.optimizer.period_range = (min_p, max_p)
            model.fit(self.time, self.flux, self.fluxErr, self.passband)

            if periods is None:
                periods, P_multi = model.periodogram_auto()
            else:
                P_multi = model.periodogram(periods)

            self.model = model
            self.best_period = model.best_period
            self.periods = periods
            self.P_multi = P_multi
            return model, periods, P_multi

        except Exception as e:
            message = '{}\nCould not run periodscan for {} at locus ID {}'.format(e, self.objectId, self.locusId)
            warnings.warn(message, RuntimeWarning)
            return None, None, None


    def phase_for_period(self, period, phase_offset=None):
        """
        Returns the phase for some period. Exposed directly to trial different
        periods, but ideally should be used by get_phase.
        """
        if phase_offset is None:
            phase_offset = 0.
        phase = ((self.time - phase_offset )/ period) % 1
        return phase


    def get_phase(self, per=None, phase_offset=None):
        """
        Returns the phase of the object
        if periodic, and best_period is set, simply returns phase with it
        if periodic, nad best_period is not set, computes a periodogram with some defaults, and returns it
        if not periodic, just returns phase relative to first event

        TODO - return phase relative to some time argument, and accept
        arguments for periodscan (maybe with *args, **kwargs)
        """
        phase = None
        if per is None:
            per = self.per # default is False
        if per:
            if self.best_period is None:
                model = getattr(self, 'model', None)
                if model is None:
                    # we don't have a model - calculate it
                    max_p = 100.
                    if self.time.max() - self.time.min() < 100:
                        max_p = self.time.max() - self.time.min()
                    model, periods, P_multi = self.periodscan(0.1, max_p)
                    if model is None:
                        # we failed to get a model - give up
                        return self.get_phase(per=None, phase_offset=phase_offset)
                    # we got a model - set the period
                    self.best_period = model.best_period
                    best_period = self.best_period
                    self.model = model
                else:
                    # we have a model - just restore the period from it
                    best_period = model.best_period
            else:
                # we have a period - just restore it
                best_period = self.best_period
            phase = self.phase_for_period(best_period, phase_offset=phase_offset)
        else:
            if phase_offset is None:
                phase_offset = 0.
            phase = self.time - phase_offset
        return phase


    def get_best_periods(self):
        model, periods, P_multi = self.periodscan(0.1, 100.)
        periods, scores = model.find_best_periods(5, return_scores=True)

        return periods, scores


